# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
import os
# noinspection PyUnresolvedReferences
# from datetime import time
import time
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
# noinspection PyUnresolvedReferences
from caffe2.perfkernels.hp_emblookup_codegen import args
from scipy.constants import alpha
from sklearn import preprocessing
from sklearn.metrics import precision_recall_fscore_support as score
from sklearn.model_selection import train_test_split
# noinspection PyUnresolvedReferences
# from tests.optimizers.test_tolerance import kwargs
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms

sns.set(color_codes=True)
# noinspection PyUnresolvedReferences
import matplotlib.pyplot as plt
# noinspection PyUnresolvedReferences
from torch.autograd import Variable
torch.manual_seed(1)
torch.cuda.manual_seed_all(1)
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.enabled = False

EPOCH = 50  # train the training data n times
TIME_STEP = None  # rnn time step
BATCH_SIZE = 218
INPUT_SIZE = 30  # rnn input size
LR = 0.01  # alerning rate

class CustomDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels

    def __len__(self):
        return (len(self.labels))

    def __getitem__(self, index):  # 通过数据样本中每个样本对应的索引，将该样本从数据列表中取出来
        return self.features[index], self.labels[index]


# dataroot="~/home/szp/guyi/CICIDS2017.csv"
dataroot = r"D:\dataset\cicfriday\mergefriday.csv"
df = pd.read_csv(dataroot, header=0, low_memory=False)
from sklearn import preprocessing

scaler = preprocessing.MinMaxScaler()  # 0-1标准化方法
# scaler = preprocessing.StandardScaler() # z-score标准化方法
df.iloc[:, 0:-1] = scaler.fit_transform(df.iloc[:, 0:-1])
# iloc[]:对数据进行位置索引，从而在数据表中提取出相应的数据
# fit_transform():先对数据拟合fit，找到整体指标，如均值、方差、最大值最小值等等（根据具体转换的目的），再对数据进行转换transform，从而实现数据的标准化、归一化
from collections import Counter

print(Counter(df[' Label']))

from sklearn.decomposition import IncrementalPCA
ipca = IncrementalPCA(n_components=INPUT_SIZE)
X = df.iloc[:, 0:-1].values
y = df[' Label'].values
ipca.fit(X, y)
X_new = ipca.transform(X)

x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=666)
x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.2, random_state=666)
# 从样本中随机的按比例选取train_data和test_data,(所要划分的样本数据集，所要划分的样本结果，样本占比，随机数的种子）

train_data = CustomDataset(x_train, y_train)
test_data = CustomDataset(x_test, y_test)
val_data = CustomDataset(x_val, y_val)
train_loader = DataLoader(train_data, batch_size=BATCH_SIZE, shuffle=True)
test_loader = DataLoader(test_data, batch_size=BATCH_SIZE, shuffle=True)
val_loader = DataLoader(val_data, batch_size=BATCH_SIZE, shuffle=True)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class LSTM(nn.Module):
    def __init__(self):
        super().__init__()
        self.lstm = nn.LSTM(
            input_size=INPUT_SIZE,
            hidden_size=32,
            #batch_first=True,
            bidirectional=False,
        )
        # 全连接层
        self.fc = nn.Linear(32, 4)

    def forward(self, x):
        # x: [batch_size, max_len, n_class]
        #BATCH_SIZE = x.shape[0]
        x = x.transpose(0, 1)  # input : [max_len, batch_size, n_class]
        # 返回一个张量，它是输入的装置版本
        # hidden_state = torch.randn(1*2, 64, 32)  # [num_layers(=1) * num_directions(=2), batch_size, n_hidden]
        # cell_state = torch.randn(1*2, 64, 32)  #[num_layers(=1) * num_direction(=2), batch_size, n_hidden]

        r_out, (h_n, h_c) = self.lstm(x)
        #r_out = self.out(r_out[:, -1, :])
        #r_out, (_, _) = self.lstm(x, (hidden_state, cell_state))  # [max_len, batch_size, n_hidden * 2]
        r_out = r_out[-1]  # [batch_size, n_hidden * 2]
        model = self.fc(r_out)  # model : [batch_size, n_class]
        return model

bilstm = LSTM().to(device=device)

print(bilstm)
print(bilstm.parameters)
# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)
optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)  # optimize all cnn parameters
# optimizer = torch.optim.SGD(bilstm.parameters(), lr=LR, weight_decay=0.001)
# torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.1, last_epoch=-1)
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted
# loss_func = torch.nn.MultiLabelMarginLoss(size_average=None, reduce=None, reduction='mean')



total = sum([param.nelement() for param in bilstm.parameters()])
print("Number of parameter: %.2fM" % (total / 1e6))

for epoch in range(EPOCH):
    print(epoch)
    epoch_start_time = time.time()
    train_acc = 0.0
    train_loss = 0.0
    val_acc = 0.0
    val_loss = 0.0

    bilstm.train()  # 确保 model 是在 train model (开启 Dropout 等...)
    for step, (b_x, b_y) in enumerate(train_loader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
        b_x = b_x.view(-1, 1, INPUT_SIZE).float()  # reshape x to (batch,time_step,input_size)
        # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
        train_pred =bilstm(b_x)# rnn output
        loss = loss_func(train_pred, b_y)  # 利用model得到预测的概率分布，这里实际上就是去呼叫model的forward函數
        optimizer.zero_grad()  # 用optimizer將model参数的gradient清零
        loss.backward()  # 利用back propagation算出每个参数的gradient
        optimizer.step()  # 以optimizer用gradient更新参数值

        train_acc += np.sum(np.argmax(train_pred.cpu().data.numpy(), axis=1) == b_y.numpy())
        train_loss += loss.item()

    bilstm.eval()
    with torch.no_grad():
        for step, (a_x, a_y) in enumerate(val_loader):
            a_x = a_x.view(-1, 1, INPUT_SIZE).float()
            val_pred = bilstm(a_x)
            loss = loss_func(val_pred, a_y)

            val_acc += np.sum(np.argmax(val_pred.cpu().data.numpy(), axis=1) == a_y.numpy())
            val_loss += loss.item()

            # 將結果 print 出來
        print('[%03d/%03d] %2.2f sec(s) Train Acc: %3.6f Loss: %3.6f | Val Acc: %3.6f loss: %3.6f' % \
            (epoch + 1, EPOCH, time.time() - epoch_start_time, \
            train_acc / train_data.__len__(), train_loss / train_data.__len__(), val_acc / val_data.__len__(),val_loss / val_data.__len__()))

import torch.quantization

quantized_model = torch.quantization.quantize_dynamic(
    bilstm, {nn.LSTM, nn.Linear}, dtype=torch.qint8
)
print(quantized_model)

def print_size_of_model(model):
    torch.save(model.state_dict(), "temp.p")
    print('Size (MB):', os.path.getsize("temp.p")/1e6)
    os.remove('temp.p')

print_size_of_model(bilstm)
print_size_of_model(quantized_model)

# import torch
# from thop import profile
#
# net = quantized_model()  # 定义好的网络模型
# input = torch.randn(132, 1, 37, 37)
# flops, params = profile(net, (inputs,))
# print('flops: ', flops, 'params: ', params)

# torch.save(bilstm, 'kztest.pkl')
# torch.save(bilstm.state_dict(), 'kztest_params.pkl')
torch.save(bilstm, 'LMtest.pth')
torch.save(bilstm.state_dict(), 'LMtest_params.pth')
torch.save(quantized_model, 'outQuant.pth')
torch.save(quantized_model.state_dict(), 'best_model_feature.pth')

# torch.jit.save(torch.jit.script(quantized_model), 'outQuant.pth')

test_transform = transforms.Compose([
    transforms.ToPILImage(),
    transforms.ToTensor(),
])

# basis = torch.jit.load('outQuant.pth')
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score, classification_report
import warnings
warnings.filterwarnings("ignore")
basis = torch.load('LMtest.pth')
basis.load_state_dict(torch.load('LMtest_params.pth'))
basis.eval()
prediction_basis = []
accuracy_list = []
precision_list = []
recall_list = []
f1_list = []
test_acc = 0.0
precision = 0.0
with torch.no_grad():
    for step, (c_x, c_y) in enumerate(test_loader):
        c_x = c_x.view(-1, 1, INPUT_SIZE).float()
        test_pred = basis(c_x)
        test_acc += np.sum(np.argmax(test_pred.cpu().data.numpy(), axis=1) == c_y.numpy())
        acc = accuracy_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1))
        accuracy_list.append(acc)
        pre_score = precision_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        precision_list.append(pre_score)
        re_score = recall_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        recall_list.append(re_score)
        f_score = f1_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        f1_list.append(f_score)
        # precision = precision+precision_split
        # recall_score += recall_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='macro')
        # f1_score += f1_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='macro')

    print('test Acc: %3.6f ' % \
          (test_acc / test_data.__len__()))
    print("accuracy_score", sum(accuracy_list)/len(accuracy_list))
    print("precision_score", sum(precision_list)/len(precision_list))
    print("recall_score", sum(recall_list)/len(recall_list))
    print("f1_score", sum(f1_list)/len(f1_list))
    # print(precision / step)
    # print('precision_score: %3.6f ' % \
    #       (precision_score / test_data.__len__()))
    # print('recall_score: %3.6f ' % \
    #       (recall_score / test_data.__len__()))
    # print('f1_score: %3.6f ' % \
    #       (f1_score / test_data.__len__()))

