# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
import os

# noinspection PyUnresolvedReferences
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
# noinspection PyUnresolvedReferences
from caffe2.perfkernels.hp_emblookup_codegen import args
from sklearn.metrics import precision_recall_fscore_support as score
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset, DataLoader

sns.set(color_codes=True)
# noinspection PyUnresolvedReferences
import matplotlib.pyplot as plt
# noinspection PyUnresolvedReferences
from torch.autograd import Variable
torch.manual_seed(1)

EPOCH = 20  # train the training data n times
TIME_STEP = None  # rnn time step
BATCH_SIZE = 64
INPUT_SIZE = 78  # rnn input size
LR = 0.01  # learning rate

class CustomDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels

    def __len__(self):
        return (len(self.labels))

    def __getitem__(self, index):  # 通过数据样本中每个样本对应的索引，将该样本从数据列表中取出来
        return self.features[index], self.labels[index]


# dataroot="~/home/szp/guyi/CICIDS2017.csv"
dataroot = r"D:\dataset\MachineLearningCVE\CICIDS2017.csv"
df = pd.read_csv(dataroot, header=0, low_memory=False)
from sklearn import preprocessing

scaler = preprocessing.MinMaxScaler()  # 0-1标准化方法
# scaler = preprocessing.StandardScaler() # z-score标准化方法
df.iloc[:, 0:-1] = scaler.fit_transform(df.iloc[:, 0:-1])
# iloc[]:对数据进行位置索引，从而在数据表中提取出相应的数据
# fit_transform():先对数据拟合fit，找到整体指标，如均值、方差、最大值最小值等等（根据具体转换的目的），再对数据进行转换transform，从而实现数据的标准化、归一化
from collections import Counter

print(Counter(df[' Label']))
x_train, x_test, y_train, y_test = train_test_split(df.iloc[:, 0:-1].values, df[' Label'].values, test_size=0.3,
                                                    random_state=42)
# 从样本中随机的按比例选取train_data和test_data,(所要划分的样本数据集，所要划分的样本结果，样本占比，随机数的种子）
train_data = CustomDataset(x_train, y_train)
test_data = CustomDataset(x_test, y_test)
trainloader = DataLoader(train_data, batch_size=BATCH_SIZE, shuffle=True)
testloader = DataLoader(test_data, batch_size=BATCH_SIZE, shuffle=False)

# PCA提取主成分
from sklearn.decomposition import PCA
pca = PCA(n_components=2, svd_solver='full')
x_train_PCA = pca.fit_transform(x_train)
x_train_PCA = pd.DataFrame(x_train_PCA)
x_train_PCA = x_train

x_test_PCA = pca.transform(x_test)
x_test_PCA = pd.DataFrame(x_test_PCA)
x_test_PCA = x_test


# class BiLSTM(nn.Module):
# def __init__(self):
# super(BiLSTM, self).__init__()
# slf.input_size = INPUT_SIZEe
# self.hidden_dim = 64 # 这其实是BiLSTM的隐藏层的特征数量，因为是双向所以是2倍，单向为2
# self.num_layers = 2
# V = 1  # vocab_size：所有不重复的词长度
# D = 6  # 由于一共有6个标签，所以embedding_dim为6
# C = 6
# self.embed = nn.Embedding(V, D, max_norm=config.max_norm)
# self.embed = nn.Embedding(V, D, padding_idx=0)
# 简单来说就是把词向量放入网络，词向量的输出应该是什么样子，输出是：(vocab_size,embedding_dim)。
# self.bilstm = nn.LSTM(D, self.hidden_dim // 2, num_layers=1, dropout=0, bidirectional=True, bias=False)
# 输入的是词向量纬度embedding_dim=6,hidden_dim的纬度为4/2=2，循环神经网络的层数为1，bidirecton代表使用双向LSTM。
# print(self.bilstm)

# self.hidden2label2 = nn.Linear(self.hidden_dim // 2, C)
# self.dropout = nn.Dropout(config.dropout)

# def forward(self, x):
# embed = self.embed(x)
# x = embed.view(len(x), embed.size(1), -1)
# bilstm_out, _ = self.bilstm(x)
# bilstm_out = torch.transpose(bilstm_out, 0, 1)
# bilstm_out = torch.transpose(bilstm_out, 1, 2)
# bilstm_out = F.tanh(bilstm_out)
# bilstm_out = F.max_pool1d(bilstm_out, bilstm_out.size(2)).squeeze(2)
# y = self.hidden2label1(bilstm_out)
# y = self.hidden2label2(y)
# logit = y
# return logit
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class BiLSTM(nn.Module):
    def __init__(self):
        super(BiLSTM, self).__init__()
        self.lstm = nn.LSTM(
            input_size=INPUT_SIZE,
            hidden_size=32,
            #batch_first=True,
            bidirectional=True,
        )
        # 全连接层
        self.fc = nn.Linear(32 * 2, 6)

    def forward(self, x):
        # x: [batch_size, max_len, n_class]
        #BATCH_SIZE = x.shape[0]
        x = x.transpose(0, 1)  # input : [max_len, batch_size, n_class]
        # 返回一个张量，它是输入的装置版本
        # hidden_state = torch.randn(1*2, 64, 32)  # [num_layers(=1) * num_directions(=2), batch_size, n_hidden]
        # cell_state = torch.randn(1*2, 64, 32)  #[num_layers(=1) * num_direction(=2), batch_size, n_hidden]

        r_out, (h_n, h_c) = self.lstm(x)
        #r_out = self.out(r_out[:, -1, :])
        #r_out, (_, _) = self.lstm(x, (hidden_state, cell_state))  # [max_len, batch_size, n_hidden * 2]
        r_out = r_out[-1]  # [batch_size, n_hidden * 2]
        model = self.fc(r_out)  # model : [batch_size, n_class]
        return model

# import torch
# import torch.nn as nn
# # noinspection PyUnresolvedReferences
# import torch.nn.functional as F
# from nni.algorithms.compression.v2.pytorch.pruning.basic_pruner import L1NormPruner
# from nni.compression.pytorch.speedup import ModelSpeedup
#
# #定义配置文件，sparsity表示稀疏度，0~1之间，数值越大表明剪枝的越多，模型越精简；
# config_list = [{
#     'sparsity': 0.5,
#     'op_types': ['Linear']
# }]
#
# model = BiLSTM()
# pruner = L1NormPruner(model, config_list)
# #基于剪枝规则获得待剪枝的通道，以masks定义；
# _, masks = pruner.compress()
#
# pruner._unwrap_model()
# ModelSpeedup(model, dummy_input=torch.ones((1,3,512,512)), masks_file=masks).speedup_model()

# import torch
# from torch import nn
# # noinspection PyUnresolvedReferences
# import torch.nn.utils.prune as prune
# # noinspection PyUnresolvedReferences
# import torch.nn.functional as F
#
# model = BiLSTM().to(device=device)
# module = model.lstm
# # print(list(module.namedmodel.lstm_parameters()))
# # print(list(module.named_buffers()))
#
# model = BiLSTM()
# parameters_to_prune = (
#     (model.lstm, 'weight_ih_l0'),
#     (model.lstm, 'weight_hh_l0'),
#     (model.lstm, 'weight_ih_l0_reverse'),
#     (model.lstm, 'weight_hh_l0_reverse'),
#     (model.fc, 'weight'),
# )
#
# prune.global_structured(
#     parameters_to_prune,
#     pruning_method=prune.L2structured,
#     amount=0.2,
# )

bilstm = BiLSTM()
print(bilstm)
print(bilstm.parameters)
optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)  # optimize all cnn parameters
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted

total = sum([param.nelement() for param in bilstm.parameters()])
print("Number of parameter: %.2fM" % (total / 1e6))
#class RNN(nn.Module):
    #def __init__(self):
       #super(RNN, self).__init__()
       #self.rnn = nn.LSTM(  # if use nn.RNN(), it hardly learns
           #input_size=INPUT_SIZE, # 输入维度
           #hidden_size=32,  # 隐藏层神经元节点个数
           #num_layers=2,  # 神经元层数
           #batch_first=True,  # input & output will has batch size as 1s dimension. e.g. (batch, time_step, input_size)
       #)
       #self.out = nn.Linear(32, 6)  # 定义全连接层

   #def forward(self, x):
        # x shape (batch, time_step, input_size)
        # r_out shape (batch, time_step, output_size)
        # h_n shape (n_layers, batch, hidden_size)
        # h_c shape (n_layers, batch, hidden_size)
       #r_out, (h_n, h_c) = self.rnn(x, None)  #  h_n就是h状态，h_c就是细胞的状态
        # choose r_out at the last time step
       #out = self.out(r_out[:, -1, :])  # 我们只要每一个time_step里的最后的一个。比如64个矩阵，每个28*28,我们只要每一个的第28次的那个数据。
       #return out

#rnn = RNN()
#print(rnn)
#optimizer = torch.optim.Adam(rnn.parameters(), lr=LR)  # optimize all cnn parameters
#loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted


# bilstm = BiLSTM()
# print(bilstm)
# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)   # optimize all cnn parameters
# loss_func = nn.CrossEntropyLoss()                       # the target label is not one-hotted

for epoch in range(EPOCH):
    for step, (b_x, b_y) in enumerate(trainloader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
        b_x = b_x.view(-1, 1, INPUT_SIZE).float()  # reshape x to (batch,time_step,input_size)
        # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
        output =bilstm(b_x)# rnn output
        loss = loss_func(output, b_y)  # cross entropy loss
        optimizer.zero_grad()  # clear gradients for this training step
        loss.backward()  # backpropagation, compute gradients
        optimizer.step()  # apply gradients
    y_predict = []
    for step, (a_x, a_y) in enumerate(testloader):
        a_x = a_x.view(-1, 1, INPUT_SIZE).float()
        test_output = bilstm(a_x)
        pred_y = torch.max(test_output, 1)[1]
        y_predict = y_predict + pred_y.numpy().tolist()
    precision, recall, fscore, support = score(y_test, y_predict)
    precision = precision * 100
    recall = recall * 100
    # print(len(precision))
    print('Epoch: ', epoch, '| train loss: %.4f' % loss.data.numpy())
    for i in range(6):
        print('Label: ', i, '| precision: %.2f' % precision[i], '%', '| recall: %.2f' % recall[i], '%',
              '| F1: %.2f' % fscore[i], '| support:', support[i])
    sum(precision)
    precision = sum(precision) / len(precision)
    sum(recall)
    recall = sum(recall) / len(recall)
    # print('average: ', '| precision: ', precision, '| recall: ', recall, '| fscore: ',fscore)
    # print('average ', '| precision: ', precision ,'| recall: ', recall)
    print('average  ', '| precision: %.2f' % precision, '%', '| recall: %.2f' % recall, '%')

# noinspection PyUnresolvedReferences
# import torch
# noinspection PyUnresolvedReferences
# import torchvision
# from torch import nn

# 保存训练好的模型
# torch.save(bilstm.state_dict(), "bilstm_method2.pth")

# 加载之前保存的模型
# model = BiLSTM()
# model.load_state_dict(torch.load("bilstm_method2.pth"))

#保存时
# torch.save(model,'bilstm_method2.pth')
#加载时
# torch.load('bilstm_method2.pth')
#
# model = torch.load("bilstm_method2.pth", map_location=torch.device('cpu'))
# print(model)
# # image = torch.reshape(image, (1, 3, 32, 32))
# model.eval()
# with torch.no_grad():
#     output = model(testloader)
# print(output)
#
# print(output.argmax(1))

# 模型测试
# model = bilstm()
# torch.save(model.state_dict(), "bilstm_method2.pth")
# # model = BiLSTM(*args, **kwargs)
# model.load_state_dict(torch.load("bilstm_method2.pth"))
# model.eval()

# for epoch in range(EPOCH):
#     y_predict = []
#     for step, (a_x, a_y) in enumerate(testloader):
#         a_x = a_x.view(-1, 1, INPUT_SIZE).float()
#         test_output = model(a_x)
#         pred_y = torch.max(test_output, 1)[1]
#         y_predict = y_predict + pred_y.numpy().tolist()
#     precision, recall, fscore, support = score(y_test, y_predict)
#     precision = precision * 100
#     recall = recall * 100
#     # print(len(precision))
#     # print('Epoch: ', epoch, '| train loss: %.4f' % loss.data.numpy())
#     for i in range(6):
#         print('Label: ', i, '| precision: %.2f' % precision[i], '%', '| recall: %.2f' % recall[i], '%',
#               '| F1: %.2f' % fscore[i], '| support:', support[i])
#     sum(precision)
#     precision = sum(precision) / len(precision)
#     sum(recall)
#     recall = sum(recall) / len(recall)
#     # print('average: ', '| precision: ', precision, '| recall: ', recall, '| fscore: ',fscore)
#     # print('average ', '| precision: ', precision ,'| recall: ', recall)
#     print('average  ', '| precision: %.2f' % precision, '%', '| recall: %.2f' % recall, '%')