# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
import os
# noinspection PyUnresolvedReferences
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
# noinspection PyUnresolvedReferences
from sklearn import preprocessing
from sklearn.metrics import precision_recall_fscore_support as score
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset, DataLoader
from cic_cnn import Pca

sns.set(color_codes=True)
# noinspection PyUnresolvedReferences
import matplotlib.pyplot as plt
# noinspection PyUnresolvedReferences
from torch.autograd import Variable
torch.manual_seed(1)

EPOCH = 1  # train the training data n times
TIME_STEP = None  # rnn time step
BATCH_SIZE = 64
INPUT_SIZE = 18  # rnn input size
LR = 0.01  # learning rate

class CustomDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels

    def __len__(self):
        return (len(self.labels))

    def __getitem__(self, index):  # 通过数据样本中每个样本对应的索引，将该样本从数据列表中取出来
        return self.features[index], self.labels[index]


# dataroot="~/home/szp/guyi/CICIDS2017.csv"
dataroot = r"D:\dataset\deeplearningCVE\MachineLearningCVE\CICIDS2017.csv"
df = pd.read_csv(dataroot, header=0, low_memory=False)
from sklearn import preprocessing

scaler = preprocessing.MinMaxScaler()  # 0-1标准化方法
# scaler = preprocessing.StandardScaler() # z-score标准化方法
df.iloc[:, 0:-1] = scaler.fit_transform(df.iloc[:, 0:-1])
# iloc[]:对数据进行位置索引，从而在数据表中提取出相应的数据
# fit_transform():先对数据拟合fit，找到整体指标，如均值、方差、最大值最小值等等（根据具体转换的目的），再对数据进行转换transform，从而实现数据的标准化、归一化
from collections import Counter

print(Counter(df[' Label']))

x_train, x_test, y_train, y_test = train_test_split(df.iloc[:, 0:-1].values, df[' Label'].values, test_size=0.2,
                                                    random_state=666)
x_train,x_test = Pca().pca(x_train,x_test)
x_train, x_val, y_train, y_val = train_test_split(x_train, y_train, test_size=0.2, random_state=666)
# 从样本中随机的按比例选取train_data和test_data,(所要划分的样本数据集，所要划分的样本结果，样本占比，随机数的种子）

train_data = CustomDataset(x_train, y_train)
test_data = CustomDataset(x_test, y_test)
val_data = CustomDataset(x_val, y_val)
train_loader = DataLoader(train_data, batch_size=BATCH_SIZE, shuffle=True)
test_loader = DataLoader(test_data, batch_size=BATCH_SIZE, shuffle=True)
val_loader = DataLoader(val_data, batch_size=BATCH_SIZE, shuffle=True)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# noinspection PyUnresolvedReferences
import optuna
# noinspection PyUnresolvedReferences
# import plotly
# import torch
from torch import nn

class BiLSTM(nn.Module):
    def __init__(self):
        super(BiLSTM, self).__init__()
        self.lstm = nn.LSTM(
            input_size=INPUT_SIZE,
            hidden_size=32,
            #batch_first=True,
            bidirectional=True,
        )
        # 全连接层
        self.fc = nn.Linear(32 * 2, 6)

    def forward(self, x):
        # x: [batch_size, max_len, n_class]
        #BATCH_SIZE = x.shape[0]
        x = x.transpose(0, 1)  # input : [max_len, batch_size, n_class]
        # 返回一个张量，它是输入的装置版本
        # hidden_state = torch.randn(1*2, 64, 32)  # [num_layers(=1) * num_directions(=2), batch_size, n_hidden]
        # cell_state = torch.randn(1*2, 64, 32)  #[num_layers(=1) * num_direction(=2), batch_size, n_hidden]
        r_out, (h_n, h_c) = self.lstm(x)
        #r_out = self.out(r_out[:, -1, :])
        #r_out, (_, _) = self.lstm(x, (hidden_state, cell_state))  # [max_len, batch_size, n_hidden * 2]
        r_out = r_out[-1]  # [batch_size, n_hidden * 2]
        model = self.fc(r_out)  # model : [batch_size, n_class]
        return model

# 定义待优化的超参数目标函数
def objective(trial):
	# 设定超参数搜索空间<1>
    batch_size = trial.suggest_int("batch_size", 64, 256, step=64)
    # batch_size = trial.suggest_int('batch_size', 4, 64)  # 为超参数指定整数取值
    # batch_size = 2 ** batch_size
    # EPOCH = trial.suggest_int('EPOCH', 10, 50)
    # flooding = trial.suggest_loguniform('flooding', 10 ** (-4), 10 ** (-2.5))
    LR = trial.suggest_loguniform('lr', 10 ** (-2.5), 10 ** (-1.5))  # 为超参数指定取值空间，且在对数域对每个值进行相同采样
    # LR = trial.suggest_float("lr", 1e-5, 1e-1, log=True)
    weight_decay = trial.suggest_loguniform('weight_decay', 10 ** (-5), 10 ** (-1))
    # network generation
    model = BiLSTM()
    weight_p, bias_p = [], []
    for name, p in model.named_parameters():
        if 'bias' in name:
            bias_p += [p]
        else:
            weight_p += [p]
    optimizer = torch.optim.Adam(model.parameters(), lr=LR, weight_decay=0.0001)
    loss_func = nn.CrossEntropyLoss()  # 定义损失函数

    for epoch in range(EPOCH):
        for step, (b_x, b_y) in enumerate(train_loader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
            b_x = b_x.view(-1, 1, INPUT_SIZE).float()  # reshape x to (batch,time_step,input_size)
            # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
            output = model(b_x)  # rnn output
            loss = loss_func(output, b_y)  # cross entropy loss
            optimizer.zero_grad()  # clear gradients for this training step
            loss.backward()  # backpropagation, compute gradients
            optimizer.step()  # apply gradients
        y_predict = []
        for step, (a_x, a_y) in enumerate(val_loader):
            a_x = a_x.view(-1, 1, INPUT_SIZE).float()
            test_output = model(a_x)
            pred_y = torch.max(test_output, 1)[1]
            y_predict = y_predict + pred_y.numpy().tolist()
        precision, recall, fscore, support = score(y_test, y_predict)
        precision = precision * 100
        recall = recall * 100
        # print(len(precision))
        print('Epoch: ', epoch, '| train loss: %.4f' % loss.data.numpy())
        for i in range(6):
            print('Label: ', i, '| precision: %.2f' % precision[i], '%', '| recall: %.2f' % recall[i], '%',
                  '| F1: %.2f' % fscore[i], '| support:', support[i])
        sum(precision)
        precision = sum(precision) / len(precision)
        sum(recall)
        recall = sum(recall) / len(recall)
        # print('average: ', '| precision: ', precision, '| recall: ', recall, '| fscore: ',fscore)
        # print('average ', '| precision: ', precision ,'| recall: ', recall)
        print('average  ', '| precision: %.2f' % precision, '%', '| recall: %.2f' % recall, '%')
    # trainer
    # trainer = SimpleTrainer(net=model, batch_size=batch_size, num_epoch=num_epoch,
    #                         optimizer=optimizer, criterion=criterion, mode='adjust',
    #                         flooding=flooding, tsb_track=None, device='cuda',
    #                         print_interval=1, early_stop=70)
    # vali_acc_track = []
    # _, _, max_vali_acc, vali_acc_track = trainer.train(x_train, x_vali, t_train, t_vali)
    # 如有需要，记录每个epoch的vali_acc<2>
    # if len(vali_acc_track):
    #     for step, vali_acc in enumerate(vali_acc_track):
    #         trial.report(vali_acc, step=step)
    return precision

# 创建优化过程<3>
study = optuna.create_study(direction='maximize')
# 开始优化
study.optimize(objective, n_trials=200)
# 可视化<4>
# importance = optuna.visualization.plot_param_importances(study)
# plt.plot.offline.plot(importance)
optuna.visualization.plot_optimization_history(study)
from optuna.visualization import plot_optimization_history
plot_optimization_history(study)
optuna.visualization.plot_param_importances(study)
