# noinspection PyUnresolvedReferences
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MultipleLocator

plt.rcParams['font.sans-serif'] = ['SimHei']

labels = ['Accuracy', 'Precision', 'Recall', 'FMeasure']
ipca = [0.9978, 0.9966, 0.9978, 0.9971]
pca = [0.9977, 0.9964, 0.9977, 0.9969]
ica = [0.9976, 0.9962, 0.9976, 0.9968]
# ipca = [99.78, 99.66, 99.78, 99.71]
# pca = [99.77, 99.64, 99.77, 99.69]
# ica = [99.76, 99.62, 99.76, 99.68]

x = np.arange(len(labels))  # the label locations
width = 0.27  # the width of the bars

fig, ax = plt.subplots()
rects1 = ax.bar(x - width - 0.02, ipca, width, label='IPCA dimensional reduction')
rects2 = ax.bar(x, pca, width, label='PCA dimensional reduction')
rects3 = ax.bar(x + width+0.02, ica, width, label='ICA dimensional reduction')

# Add some text for labels, title and custom x-axis tick labels, etc.
ax.set_xlabel('Metric')
ax.set_ylabel('Scores')
ax.set_title('Comparison of dimensionality reduction algorithms')
ax.set_xticks(x)
ax.set_xticklabels(labels)
ax.legend()
plt.ylim((0.99, 1.00))
y_major_locator = MultipleLocator(0.002)#以每3显示
ax.yaxis.set_major_locator(y_major_locator)


def autolabel(rects):
    """Attach a text label above each bar in *rects*, displaying its height."""
    for rect in rects:
        height = rect.get_height()
        ax.annotate('{}'.format(height),
                    xy=(rect.get_x() + rect.get_width() / 2, height),
                    xytext=(0, 3),  # 3 points vertical offset
                    textcoords="offset points",
                    ha='center', va='bottom')


autolabel(rects1)
autolabel(rects2)
autolabel(rects3)

fig.tight_layout()

plt.show()
