import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import torch
import time
import os
from sklearn.model_selection import train_test_split, KFold, cross_val_score
# from sklearn.cross_validation import train_test_split
from torch.utils.data import Dataset, DataLoader
from multipre import x_train, x_val, x_test, y_train, y_val, y_test
from sklearn.linear_model import LogisticRegression
# from sklearn.cross_validation import KFold, cross_val_score
from sklearn.metrics import confusion_matrix,recall_score,classification_report,f1_score
from sklearn.ensemble import IsolationForest,RandomForestClassifier
from sklearn import svm
import itertools
from torch import nn
from torchvision import transforms

# noinspection PyUnresolvedReferences
import matplotlib.pyplot as plt
# noinspection PyUnresolvedReferences
from torch.autograd import Variable
torch.manual_seed(1)
np.random.seed(1)
torch.manual_seed(1)
torch.cuda.manual_seed_all(1)
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.enabled = False

EPOCH = 60  # train the training data n times
TIME_STEP = None  # rnn time step
BATCH_SIZE = 292
INPUT_SIZE = 24  # rnn input size
LR = 0.001  # learning rate

class CustomDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels

    def __len__(self):
        return (len(self.labels))

    def __getitem__(self, index):  # 通过数据样本中每个样本对应的索引，将该样本从数据列表中取出来
        return self.features[index], self.labels[index]


train_data = CustomDataset(x_train, y_train)
test_data = CustomDataset(x_test, y_test)
val_data = CustomDataset(x_val, y_val)
train_loader = DataLoader(train_data, batch_size=BATCH_SIZE, shuffle=True)
test_loader = DataLoader(test_data, batch_size=BATCH_SIZE, shuffle=True)
val_loader = DataLoader(val_data, batch_size=BATCH_SIZE, shuffle=True)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class ANNModel(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim):
        super(ANNModel, self).__init__()
        # 定义层
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.relu1 = nn.ReLU()  # nn.Linear为线性关系，加上激活函数转为非线性

        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.relu2 = nn.ReLU()

        self.out = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        out = self.fc1(x)
        out = self.relu1(out)
        out = self.fc2(out)
        out = self.relu2(out)
        out = self.out(out[:, -1, :])
        return out

input_dim = 24  # 一维向量长度
hidden_dim = 64  # hidden layer 神经元个数
output_dim = 6  # 10个类

bilstm = ANNModel(input_dim, hidden_dim, output_dim)

print(bilstm)
print(bilstm.parameters)
# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)
optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)  # optimize all cnn parameters
# optimizer = torch.optim.SGD(bilstm.parameters(), lr=LR, weight_decay=0.001)
# torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.1, last_epoch=-1)
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted

# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR, weight_decay=0.004)  # optimize all cnn parameters
# torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.1, last_epoch=-1)
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
# loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted
total = sum([param.nelement() for param in bilstm.parameters()])
print("Number of parameter: %.2fM" % (total / 1e6))

for epoch in range(EPOCH):
    print(epoch)
    epoch_start_time = time.time()
    train_acc = 0.0
    train_loss = 0.0
    val_acc = 0.0
    val_loss = 0.0

    bilstm.train()  # 确保 model 是在 train model (开启 Dropout 等...)
    for step, (b_x, b_y) in enumerate(train_loader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
        b_x = b_x.view(-1, 1, INPUT_SIZE).float()  # reshape x to (batch,time_step,input_size)
        # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
        train_pred =bilstm(b_x)# rnn output
        loss = loss_func(train_pred, b_y)  # 利用model得到预测的概率分布，这里实际上就是去呼叫model的forward函數
        optimizer.zero_grad()  # 用optimizer將model参数的gradient清零
        loss.backward()  # 利用back propagation算出每个参数的gradient
        optimizer.step()  # 以optimizer用gradient更新参数值

        train_acc += np.sum(np.argmax(train_pred.cpu().data.numpy(), axis=1) == b_y.numpy())
        train_loss += loss.item()

    bilstm.eval()
    with torch.no_grad():
        for step, (a_x, a_y) in enumerate(val_loader):
            a_x = a_x.view(-1, 1, INPUT_SIZE).float()
            val_pred = bilstm(a_x)
            loss = loss_func(val_pred, a_y)

            val_acc += np.sum(np.argmax(val_pred.cpu().data.numpy(), axis=1) == a_y.numpy())
            val_loss += loss.item()

            # 將結果 print 出來
        print('[%03d/%03d] %2.2f sec(s) Train Acc: %3.6f Loss: %3.6f | Val Acc: %3.6f loss: %3.6f' % \
            (epoch + 1, EPOCH, time.time() - epoch_start_time, \
            train_acc / train_data.__len__(), train_loss / train_data.__len__(), val_acc / val_data.__len__(),val_loss / val_data.__len__()))

import torch.quantization

quantized_model = torch.quantization.quantize_dynamic(
    bilstm, {nn.LSTM}, dtype=torch.qint8
)
print(quantized_model)

def print_size_of_model(model):
    torch.save(model.state_dict(), "temp.p")
    print('Size (MB):', os.path.getsize("temp.p")/1e6)
    os.remove('temp.p')

print_size_of_model(bilstm)
print_size_of_model(quantized_model)

torch.save(bilstm, 'BMtest.pth')
torch.save(bilstm.state_dict(), 'BMtest_params.pth')
torch.save(quantized_model, 'outQuant.pth')
torch.save(quantized_model.state_dict(), 'best_model_feature.pth')

# torch.save(bilstm.state_dict(), 'best_model_feature.pth')
# torch.save(quantized_model, 'outQuant.pth')
# torch.save(quantized_model.state_dict(), 'best_model_feature.pth')

# torch.jit.save(torch.jit.script(quantized_model), 'outQuant.pth')

# test_transform = transforms.Compose([
#     transforms.ToPILImage(),
#     transforms.ToTensor(),
# ])
#
# torch.save(bilstm, 'kztest.pkl')
# torch.save(bilstm.state_dict(), 'kztest_params.pkl')
#
torch.jit.save(torch.jit.script(quantized_model), 'outQuant.pth')

# testing 時不需做 data augmentation
test_transform = transforms.Compose([
    transforms.ToPILImage(),
    transforms.ToTensor(),
])

def confusion_matrix(preds, labels, conf_matrix):
    preds = torch.argmax(preds, 1)
    for p, t in zip(preds, labels):
        conf_matrix[p, t] += 1
    return conf_matrix

# basis = torch.load('kzqt.pkl')
# basis.load_state_dict(torch.load('kzqt_params.pkl'))
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score, classification_report
import warnings
warnings.filterwarnings("ignore")
# basis = torch.jit.load('outQuant.pth')
basis = torch.load('BMtest.pth')
basis.load_state_dict(torch.load('BMtest_params.pth'))
basis.eval()
prediction_basis = []
accuracy_list = []
precision_list = []
recall_list = []
f1_list = []
test_acc = 0.0
precision = 0.0
conf_matrix = torch.zeros(6, 6)
with torch.no_grad():
    for step, (c_x, c_y) in enumerate(test_loader):
        c_x = c_x.view(-1, 1, INPUT_SIZE).float()
        test_pred = basis(c_x)
        test_acc += np.sum(np.argmax(test_pred.cpu().data.numpy(), axis=1) == c_y.numpy())
        acc = accuracy_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1))
        accuracy_list.append(acc)
        pre_score = precision_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        precision_list.append(pre_score)
        re_score = recall_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        recall_list.append(re_score)
        f_score = f1_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        f1_list.append(f_score)
        conf_matrix = confusion_matrix(test_pred, c_y, conf_matrix)
        # conf_matrix = conf_matrix.cpu()
        # conf_matrix = np.array(conf_matrix.cpu())  # 将混淆矩阵从gpu转到cpu再转到np
        corrects = conf_matrix.diagonal(offset=0)  # 抽取对角线的每种分类的识别正确个数
        per_kinds = conf_matrix.sum(axis=1)  # 抽取每个分类数据总的测试条数

    print(conf_matrix)

    # 获取每种Emotion的识别准确率
    print("每种情感总个数：", per_kinds)
    print("每种情感预测正确的个数：", corrects)
    print("每种情感的识别准确率为：{0}".format([rate * 100 for rate in corrects / per_kinds]))
    print('test Acc: %3.6f ' % \
          (test_acc / test_data.__len__()))
    print("accuracy_score", sum(accuracy_list) / len(accuracy_list))
    print("precision_score", sum(precision_list) / len(precision_list))
    print("recall_score", sum(recall_list) / len(recall_list))
    print("f1_score", sum(f1_list) / len(f1_list))
# 绘制混淆矩阵
Emotion = 6  # 这个数值是具体的分类数，大家可以自行修改
labels = ['BENIGN', 'ACK', 'Scan', 'SYN', 'UDP', 'UDPP']  # 每种类别的标签

# 显示数据
plt.imshow(conf_matrix, cmap=plt.cm.Blues)

# 在图中标注数量/概率信息
thresh = conf_matrix.max() / 2  # 数值颜色阈值，如果数值超过这个，就颜色加深。
for x in range(6):
    for y in range(6):
        # 注意这里的matrix[y, x]不是matrix[x, y]
        info = int(conf_matrix[y, x])
        plt.text(x, y, info,
                 verticalalignment='center',
                 horizontalalignment='center',
                 color="white" if info > thresh else "black")

plt.grid(False)  # 关闭背景的网格线
plt.tight_layout()  # 保证图不重叠
plt.yticks(range(6), labels, fontsize=12)
plt.xticks(range(6), labels, fontsize=12, rotation=45)  # X轴字体倾斜45°

plt.ylabel('Target', fontsize=16)
plt.xlabel('Predicted', fontsize=16)
plt.show()
# plt.close()