import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import torch
import time
from sklearn.model_selection import train_test_split, KFold, cross_val_score
# from sklearn.cross_validation import train_test_split
from torch.utils.data import Dataset, DataLoader
from pre import x_train, x_val, x_test, y_train, y_val, y_test
from sklearn.linear_model import LogisticRegression
# from sklearn.cross_validation import KFold, cross_val_score
from sklearn.metrics import confusion_matrix,recall_score,classification_report,f1_score
from sklearn.ensemble import IsolationForest,RandomForestClassifier
from sklearn import svm
import itertools
from torch import nn

# noinspection PyUnresolvedReferences
import matplotlib.pyplot as plt
# noinspection PyUnresolvedReferences
from torch.autograd import Variable
torch.manual_seed(1)

EPOCH = 70  # train the training data n times
TIME_STEP = None  # rnn time step
BATCH_SIZE = 256
INPUT_SIZE = 28  # rnn input size
LR = 0.03  # learning rate


class CustomDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels

    def __len__(self):
        return (len(self.labels))

    def __getitem__(self, index):  # 通过数据样本中每个样本对应的索引，将该样本从数据列表中取出来
        return self.features[index], self.labels[index]


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


class BiLSTM(nn.Module):
    def __init__(self):
        super(BiLSTM, self).__init__()
        self.lstm = nn.LSTM(
            input_size=INPUT_SIZE,
            hidden_size=32,
            #batch_first=True,
            bidirectional=True,
        )
        # 全连接层
        self.fc = nn.Linear(32 * 2, 2)

    def forward(self, x):
        # x: [batch_size, max_len, n_class]
        #BATCH_SIZE = x.shape[0]
        x = x.transpose(0, 1)  # input : [max_len, batch_size, n_class]
        # 返回一个张量，它是输入的装置版本
        # hidden_state = torch.randn(1*2, 64, 32)  # [num_layers(=1) * num_directions(=2), batch_size, n_hidden]
        # cell_state = torch.randn(1*2, 64, 32)  #[num_layers(=1) * num_direction(=2), batch_size, n_hidden]

        r_out, (h_n, h_c) = self.lstm(x)
        #r_out = self.out(r_out[:, -1, :])
        #r_out, (_, _) = self.lstm(x, (hidden_state, cell_state))  # [max_len, batch_size, n_hidden * 2]
        r_out = r_out[-1]  # [batch_size, n_hidden * 2]
        model = self.fc(r_out)  # model : [batch_size, n_class]
        return model


# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR, weight_decay=0.004)  # optimize all cnn parameters
# torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.1, last_epoch=-1)
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
# loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted

kf = KFold(n_splits=5, shuffle=True)
acc_list = []
pre_list = []
recall_list = []
f1_list = []
for train_index, test_index in kf.split(x_train):
    # net = EdgeNeXt(in_channels=1, stem_features=64, depths=[1, 1, 3, 1], widths=[256, 512, 1024, 2048], num_classes=5,
    #                classifier_dropout=0.5).to(device)
    criterion = nn.CrossEntropyLoss()
    # optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR, weight_decay=0.007)
    # optimizer = torch.optim.Adam(lr=1e-3, params=bilstm.parameters())
    bilstm = BiLSTM()
    optimizer = torch.optim.Adam(lr=0.01, params=bilstm.parameters())

    # X_train, X_val = x_train[train_index, :], x_train[test_index, :]
    # Y_train, Y_val = y_train[train_index], y_train[test_index]
    train_data = CustomDataset(x_train, y_train)
    # train_data = CustomDataset(X_train, Y_train)
    train_loader = DataLoader(train_data, batch_size=BATCH_SIZE, shuffle=True)
    val_data = CustomDataset(x_val, y_val)
    # val_data = CustomDataset(X_val, Y_val)
    val_loader = DataLoader(val_data, batch_size=BATCH_SIZE, shuffle=True)

    for epoch in range(EPOCH):
        # print(epoch)
        epoch_start_time = time.time()
        train_acc = 0.0
        train_loss = 0.0
        val_acc = 0.0
        val_loss = 0.0

        bilstm.train()  # 确保 model 是在 train model (开启 Dropout 等...)
        for step, (b_x, b_y) in enumerate(train_loader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
            b_x = b_x.view(-1, 1, INPUT_SIZE).float()  # reshape x to (batch,time_step,input_size)
            # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
            train_pred =bilstm(b_x)# rnn output
            loss = criterion(train_pred, b_y)  # 利用model得到预测的概率分布，这里实际上就是去呼叫model的forward函數
            optimizer.zero_grad()  # 用optimizer將model参数的gradient清零
            loss.backward()  # 利用back propagation算出每个参数的gradient
            optimizer.step()  # 以optimizer用gradient更新参数值

            train_acc += np.sum(np.argmax(train_pred.cpu().data.numpy(), axis=1) == b_y.numpy())
            train_loss += loss.item()

        bilstm.eval()
        with torch.no_grad():
            for step, (a_x, a_y) in enumerate(val_loader):
                a_x = a_x.view(-1, 1, INPUT_SIZE).float()
                val_pred = bilstm(a_x)
                loss = criterion(val_pred, a_y)

                val_acc += np.sum(np.argmax(val_pred.cpu().data.numpy(), axis=1) == a_y.numpy())
                val_loss += loss.item()

                # 將結果 print 出來
            print('[%03d/%03d] %2.2f sec(s) Train Acc: %3.6f Loss: %3.6f | Val Acc: %3.6f loss: %3.6f' % \
                (epoch + 1, EPOCH, time.time() - epoch_start_time, \
                train_acc / train_data.__len__(), train_loss / train_data.__len__(), val_acc / val_data.__len__(),val_loss / val_data.__len__()))

# for epoch in range(EPOCH):
#     print(epoch)
#     epoch_start_time = time.time()
#     train_acc = 0.0
#     train_loss = 0.0
#     val_acc = 0.0
#     val_loss = 0.0
#
#     bilstm.train()  # 确保 model 是在 train model (开启 Dropout 等...)
#     for batch_idx, (data, target, _) in enumerate(train_loader):
#         data, target = data.to(device), target.to(device)
#         # data = torch.reshape(data, shape=(data.size(0), 1, 14, 14))
#         data = torch.unsqueeze(data, dim=1)
#     # for step, (b_x, b_y) in enumerate(train_loader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
#     #     data = data.view(-1, 1, INPUT_SIZE).float()  # reshape x to (batch,time_step,input_size)
#         # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
#         train_pred =bilstm(data)# rnn output
#         loss = loss_func(train_pred, target.long())  # 利用model得到预测的概率分布，这里实际上就是去呼叫model的forward函數
#         optimizer.zero_grad()  # 用optimizer將model参数的gradient清零
#         loss.backward()  # 利用back propagation算出每个参数的gradient
#         optimizer.step()  # 以optimizer用gradient更新参数值
#
#         train_acc += np.sum(np.argmax(train_pred.cpu().data.numpy(), axis=1) == target.numpy())
#         train_loss += loss.item()
#
#     bilstm.eval()
#     with torch.no_grad():
#         for batch_idx, (data, target, input_indices) in enumerate(val_loader):
#             data, target = data.to(device), target.to(device)
#             data = torch.unsqueeze(data, dim=1)
#             val_pred = bilstm(data)
#             loss = loss_func(val_pred, target.long())
#
#             val_acc += np.sum(np.argmax(val_pred.cpu().data.numpy(), axis=1) == target.numpy())
#             val_loss += loss.item()
#
#             # 將結果 print 出來
#         print('[%03d/%03d] %2.2f sec(s) Train Acc: %3.6f Loss: %3.6f | Val Acc: %3.6f loss: %3.6f' % \
#             (epoch + 1, EPOCH, time.time() - epoch_start_time, \
#             train_acc / train_data.__len__(), train_loss / train_data.__len__(), val_acc / val_data.__len__(),val_loss / val_data.__len__()))


# print("\nImporting to main.py ...##")
# from pre import Xtrain,Xtest,Ytrain,Ytest
# from sklearn.metrics import accuracy_score
# from matplotlib import pyplot as plt
# print("\n\n\tRunning main.py\n")
#
# #Training
# print("\n\tRunning Logistic Regression . . . \n")
# from sklearn.linear_model import LogisticRegression
# model1=LogisticRegression(solver='lbfgs',max_iter=1000)
# model1.fit(Xtrain,Ytrain)
# prediction1=model1.predict(Xtest)
# print(model1.score(Xtest,Ytest)*100)
# print("Accuracy Score MODEL 1 is : ",accuracy_score(prediction1,Ytest))
#
#
# #Training
# print("\n\tRunning Preceptron . . . \n")
# from sklearn.linear_model import Perceptron
# model2=Perceptron(eta0=0.2,max_iter=1000,tol=1e-3,verbose=0,early_stopping = True ,validation_fraction=0.1)
# model2.fit(Xtrain,Ytrain)
# prediction2=model2.predict(Xtest)
# print(model2.score(Xtest,Ytest)*100)
# print("Accuracy Score MODEL 2 is : ",accuracy_score(prediction2,Ytest))
#
#
# #Training
# print("\n\tRunning DecisionTreeRegressor . . . \n")
# from sklearn.tree import DecisionTreeRegressor
# model3=DecisionTreeRegressor()
# model3.fit(Xtrain,Ytrain)
# prediction3=model3.predict(Xtest)
# print(model3.score(Xtest,Ytest)*100)
# #print("Accuracy Score MODEL 3 is : ",accuracy_score(prediction3,Ytest))
# #Accuracy cannot be shown as this regression model
#
# #Training
# print("\n\tRunning Gaussian Naive Bayes . . . \n")
# from sklearn.naive_bayes import GaussianNB
# model4=GaussianNB()
# model4.fit(Xtrain,Ytrain)
# prediction4=model4.predict(Xtest)
# print(model4.score(Xtest,Ytest)*100)
# print("Accuracy Score MODEL 4 is : ",accuracy_score(prediction4,Ytest))
#
#
# #Training
# print("\n\tRunning KNN Classifier . . .\n")
# from sklearn.neighbors import KNeighborsClassifier
# model5=GaussianNB()
# model5.fit(Xtrain,Ytrain)
# prediction5=model5.predict(Xtest)
# print(model5.score(Xtest,Ytest)*100)
# print("Accuracy Score MODEL 4 is : ",accuracy_score(prediction5,Ytest))
#
# Ytest_p=Ytest[:100]
# prediction1_p=prediction1[:100]
# plt.xlabel('X(Time->)')
# plt.ylabel('0 For Clean Traffic(LOW) and Threat Traffic as 1(HIGH)')
# plt.plot(Ytest_p,c='b',label="Test Data")
# plt.plot(prediction1_p,c='r',label="Predicted Attack")
# plt.legend(loc='upper left')
# plt.show()

