import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import torch
import time
import os
from sklearn.model_selection import train_test_split, KFold, cross_val_score
# from sklearn.cross_validation import train_test_split
from torch.utils.data import Dataset, DataLoader
from multipre import x_train, x_val, x_test, y_train, y_val, y_test
from sklearn.linear_model import LogisticRegression
# from sklearn.cross_validation import KFold, cross_val_score
from sklearn.metrics import confusion_matrix,recall_score,classification_report,f1_score
from sklearn.ensemble import IsolationForest,RandomForestClassifier
from sklearn import svm
import itertools
from torch import nn
from torchvision import transforms

# noinspection PyUnresolvedReferences
import matplotlib.pyplot as plt
# noinspection PyUnresolvedReferences
from torch.autograd import Variable
torch.manual_seed(1)
np.random.seed(1)
torch.manual_seed(1)
torch.cuda.manual_seed_all(1)
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.enabled = False

EPOCH = 60  # train the training data n times
TIME_STEP = None  # rnn time step
BATCH_SIZE = 292
INPUT_SIZE = 25  # rnn input size
LR = 0.001  # learning rate

class CustomDataset(Dataset):
    def __init__(self, features, labels):
        self.features = features
        self.labels = labels

    def __len__(self):
        return (len(self.labels))

    def __getitem__(self, index):  # 通过数据样本中每个样本对应的索引，将该样本从数据列表中取出来
        return self.features[index], self.labels[index]


train_data = CustomDataset(x_train, y_train)
test_data = CustomDataset(x_test, y_test)
val_data = CustomDataset(x_val, y_val)
train_loader = DataLoader(train_data, batch_size=BATCH_SIZE, shuffle=True)
test_loader = DataLoader(test_data, batch_size=BATCH_SIZE, shuffle=True)
val_loader = DataLoader(val_data, batch_size=BATCH_SIZE, shuffle=True)

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class CNN(nn.Module):
    def __init__(self):

        super(CNN, self).__init__()
        self.conv1 = nn.Conv2d(1, 32, kernel_size=5)
        self.bn1 = nn.BatchNorm2d(32)
        self.relu = nn.ReLU(inplace=True)

        self.conv2 = nn.Conv2d(32, 64, kernel_size=1)
        self.bn2 = nn.BatchNorm2d(64)

        self.conv3 = nn.Conv2d(64, 128, kernel_size=1)
        self.bn3 = nn.BatchNorm2d(128)

        self.fc = nn.Linear(128, 6)

    def forward(self, x):
        x = self.relu(self.bn1(self.conv1(x)))
        x = self.relu(self.bn2(self.conv2(x)))
        x = self.relu(self.bn3(self.conv3(x)))
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


bilstm = CNN().to(device=device)

print(bilstm)
print(bilstm.parameters)
# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)
optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR)  # optimize all cnn parameters
# optimizer = torch.optim.SGD(bilstm.parameters(), lr=LR, weight_decay=0.001)
# torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.1, last_epoch=-1)
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted

# optimizer = torch.optim.Adam(bilstm.parameters(), lr=LR, weight_decay=0.004)  # optimize all cnn parameters
# torch.optim.lr_scheduler.StepLR(optimizer, step_size=20, gamma=0.1, last_epoch=-1)
# optimizer = torch.optim.RMSprop(bilstm.parameters(), lr=LR)
# loss_func = nn.CrossEntropyLoss()  # the target label is not one-hotted
total = sum([param.nelement() for param in bilstm.parameters()])
print("Number of parameter: %.2fM" % (total / 1e6))

for epoch in range(EPOCH):
    print(epoch)
    epoch_start_time = time.time()
    train_acc = 0.0
    train_loss = 0.0
    val_acc = 0.0
    val_loss = 0.0

    bilstm.train()  # 确保 model 是在 train model (开启 Dropout 等...)
    for step, (b_x, b_y) in enumerate(train_loader):  # (b_x, b_y)里面包含输入数据（inputs）(tensor类型的）和标签（labels）(tensor类型）
        b_x = torch.reshape(b_x, shape=(b_x.size(0), 1, 5, 5))
        # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
        b_x = b_x.to(torch.float32)
        # 重新定义矩阵的形状，view中一个参数定为-1，代表动态调整这个维度上的元素个数，以保证元素的总数不变。
        train_pred =bilstm(b_x)# rnn output
        loss = loss_func(train_pred, b_y)  # 利用model得到预测的概率分布，这里实际上就是去呼叫model的forward函數
        optimizer.zero_grad()  # 用optimizer將model参数的gradient清零
        loss.backward()  # 利用back propagation算出每个参数的gradient
        optimizer.step()  # 以optimizer用gradient更新参数值

        train_acc += np.sum(np.argmax(train_pred.cpu().data.numpy(), axis=1) == b_y.numpy())
        train_loss += loss.item()

    bilstm.eval()
    with torch.no_grad():
        for step, (a_x, a_y) in enumerate(val_loader):
            a_x = torch.reshape(a_x, shape=(a_x.size(0), 1, 5, 5))
            a_x = a_x.to(torch.float32)
            val_pred = bilstm(a_x)
            loss = loss_func(val_pred, a_y)

            val_acc += np.sum(np.argmax(val_pred.cpu().data.numpy(), axis=1) == a_y.numpy())
            val_loss += loss.item()

            # 將結果 print 出來
        print('[%03d/%03d] %2.2f sec(s) Train Acc: %3.6f Loss: %3.6f | Val Acc: %3.6f loss: %3.6f' % \
            (epoch + 1, EPOCH, time.time() - epoch_start_time, \
            train_acc / train_data.__len__(), train_loss / train_data.__len__(), val_acc / val_data.__len__(),val_loss / val_data.__len__()))

import torch.quantization

quantized_model = torch.quantization.quantize_dynamic(
    bilstm, {nn.LSTM, nn.Linear}, dtype=torch.qint8
)
print(quantized_model)

def print_size_of_model(model):
    torch.save(model.state_dict(), "temp.p")
    print('Size (MB):', os.path.getsize("temp.p")/1e6)
    os.remove('temp.p')

print_size_of_model(bilstm)
print_size_of_model(quantized_model)

torch.save(bilstm, 'cnnM.pth')
torch.save(bilstm.state_dict(), 'cnn_params.pth')
torch.save(quantized_model, 'outQuant.pth')
torch.save(quantized_model.state_dict(), 'best_model_feature.pth')


torch.jit.save(torch.jit.script(quantized_model), 'outQuant.pth')

# testing 時不需做 data augmentation
test_transform = transforms.Compose([
    transforms.ToPILImage(),
    transforms.ToTensor(),
])

from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score, classification_report
import warnings
warnings.filterwarnings("ignore")
# basis = torch.jit.load('outQuant.pth')
basis = torch.load('cnnM.pth')
basis.load_state_dict(torch.load('cnn_params.pth'))
basis.eval()
prediction_basis = []
accuracy_list = []
precision_list = []
recall_list = []
f1_list = []
test_acc = 0.0
precision = 0.0
with torch.no_grad():
    for step, (c_x, c_y) in enumerate(test_loader):
        c_x = torch.reshape(c_x, shape=(c_x.size(0), 1, 5, 5))
        c_x = c_x.to(torch.float32)
        test_pred = basis(c_x)
        test_acc += np.sum(np.argmax(test_pred.cpu().data.numpy(), axis=1) == c_y.numpy())
        acc = accuracy_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1))
        accuracy_list.append(acc)
        pre_score = precision_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        precision_list.append(pre_score)
        re_score = recall_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        recall_list.append(re_score)
        f_score = f1_score(c_y, np.argmax(test_pred.cpu().data.numpy(), axis=1), average='weighted')
        f1_list.append(f_score)

    print('test Acc: %3.6f ' % \
          (test_acc / test_data.__len__()))
    print("accuracy_score", sum(accuracy_list) / len(accuracy_list))
    print("precision_score", sum(precision_list) / len(precision_list))
    print("recall_score", sum(recall_list) / len(recall_list))
    print("f1_score", sum(f1_list) / len(f1_list))
