# noinspection PyUnresolvedReferences
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MultipleLocator

plt.rcParams['font.sans-serif'] = ['SimHei']

labels = ['BENIGN', 'ACK', 'Scan', 'SYN', 'UDP', 'UDPP']
moxin1 = [0.9993, 0.9992, 0.9999, 1.0000, 1.0000, 1.0000]
moxin2 = [0.9990, 0.9581, 0.9999, 1.0000, 0.9971, 0.9978]
moxin3 = [0.9990, 0.9134, 1.0000, 1.0000, 0.9933, 1.0000]
moxin4 = [0.9990, 0.9548, 1.0000, 1.0000, 0.9968, 1.0000]
# ipca = [99.78, 99.66, 99.78, 99.71]
# pca = [99.77, 99.64, 99.77, 99.69]
# ica = [99.76, 99.62, 99.76, 99.68]

x = np.arange(len(labels))  # the label locations
width = 0.2  # the width of the bars

fig, ax = plt.subplots()
rects1 = ax.bar(x - 2 * width + 0.1, moxin1, width, label='DL-BiLSTM')
rects2 = ax.bar(x - width + 0.1, moxin2, width, label='DNN-LSTM')
rects3 = ax.bar(x + width - 0.1, moxin3, width, label='PCA+LSTM')
rects4 = ax.bar(x + 2 * width - 0.1, moxin4, width, label='BiLSTM')

# Add some text for labels, title and custom x-axis tick labels, etc.
ax.set_xlabel('Category')
ax.set_ylabel('Accuracy')
ax.set_title('Comparison of detection accuracy of various types')
ax.set_xticks(x)
ax.set_xticklabels(labels)
ax.legend()
plt.ylim((0.91, 1.01))
y_major_locator = MultipleLocator(0.01)  # 以每3显示
ax.yaxis.set_major_locator(y_major_locator)


def autolabel(rects):
    """Attach a text label above each bar in *rects*, displaying its height."""
    for rect in rects:
        height = rect.get_height()
        ax.annotate('{}'.format(height),
                    xy=(rect.get_x() + rect.get_width() / 2, height),
                    xytext=(0, 3),  # 3 points vertical offset
                    textcoords="offset points",
                    ha='center', va='bottom')

# autolabel(rects1)
# autolabel(rects2)
# autolabel(rects3)

fig.tight_layout()

plt.show()
